"""Generates tie-break points on tied players.

Tie-breaks supported:
  * Direct Encounter
  * Number of wins
  * Sonneborn-Berger
"""


from typing import List, Dict

import pandas as pd
import pgnhelper.utility


def num_wins(result_df: pd.DataFrame, ranking_df: pd.DataFrame) -> pd.DataFrame:
    """Creates a dataframe with Win column.
    
    If a game has an armageddon tie-break, we will only count the number of wins
    based from the normal game only.
    """
    ret = ranking_df.copy()
    players = list(ret.Name)
    tb = {}
    for _, g in ret.groupby(['Score']):
        if len(g) > 1:
            for p in g.Name:
                df_w = result_df.loc[(result_df.White == p) & (result_df.Result == '1-0') & (result_df.Arm == 0)]
                df_b = result_df.loc[(result_df.Black == p) & (result_df.Result == '0-1') & (result_df.Arm == 0)]
                num_wins = len(df_w) + len(df_b)
                tb.update({p: num_wins})

    # Create new column Wins.
    wins = []
    for p in players:
        if p in tb:
            wins.append(tb[p])
        else:
            wins.append(0)
    ret['Wins'] = wins
    return ret


def direct_encounter(result_df: pd.DataFrame, ranking_df: pd.DataFrame, winpoint=1.0, drawpoint=0.5, winpointarm=1.0, losspointarm=0.0) -> pd.DataFrame:
    """Creates a dataframe with DE column or direct encounter.
    """
    players = list(ranking_df.Name)
    tb = {}
    ret: pd.DataFrame = ranking_df.copy()
    for _, g in ret.groupby(['Score']):
        if len(g) > 1:
            for p in g.Name:
                s = 0
                for op in g.Name:
                    if p == op:
                        continue
                    score = pgnhelper.utility.get_encounter_score(result_df, p, op, winpoint, drawpoint, winpointarm, losspointarm)
                    s += score[0]
                    tb.update({p: {op: s}})

    # Create new column DE.
    de = []
    for p in players:
        if p in tb:
            s = 0
            for k, v in tb.items():
                if k == p:
                    for _, v1 in v.items():
                        s += v1
            de.append(s)
        else:
            de.append(0)
    ret['DE'] = de
    return ret


def sonneborn_berger(result_df: pd.DataFrame, ranking_df: pd.DataFrame,
        gpe: int=1, winpoint=1.0, drawpoint=0.5) -> pd.DataFrame:
    """Creates a dataframe with SB column for Sonneborn-Berger score.

    Args:
      result_df: A dataframe of [Round, White, Black, Result].
      ranking_df: A dataframe of standing, [Name, Games, Score].
      gpe: games per encounter

    Returns:
      A dataframe of round-robin result table.
    """
    tb: Dict[str, int] = {}
    ret: pd.DataFrame = ranking_df.copy()
    players = list(ret.Name)

    # 1. Loop thru the tied players.
    for _, g in ret.groupby(['Score']):
        if len(g) > 1:
            for p in g['Name']:
                tb_score = 0
                for m in players:
                    if p == m:
                        continue
                    match_score = 0
            
                    # 2. Get the score when player wins or draws.
                    df_ww = result_df.loc[(result_df.White == p) & (result_df.Black == m) & (result_df.Result == '1-0')]
                    df_wd = result_df.loc[(result_df.White == p) & (result_df.Black == m) & (result_df.Result == '1/2-1/2')]
                    df_bw = result_df.loc[(result_df.Black == p) & (result_df.White == m) & (result_df.Result == '0-1')]
                    df_bd = result_df.loc[(result_df.Black == p) & (result_df.White == m) & (result_df.Result == '1/2-1/2')]

                    # 3. Calculate the scores.
                    match_score += winpoint * len(df_ww) + winpoint * len(df_bw) + len(df_wd) * drawpoint + len(df_bd) * drawpoint

                    if match_score > drawpoint * gpe:
                        tb_score += ret.loc[ret.Name == m].Score.iloc[0]
                    elif match_score == drawpoint * gpe:
                        tb_score += ret.loc[ret.Name == m].Score.iloc[0] / 2

                tb.update({p: tb_score})

    # 4. Create new column SB.
    tb_sb: List = []
    for p in players:
        if p not in tb:
            tb_sb.append(0)
            continue
        tb_sb.append(tb[p])
    ret['SB'] = tb_sb
    return ret



"""Sort games by game tags.
"""


from typing import List
from operator import attrgetter


class Game:
    """Manages sorting of games.
    """
    def __init__(self):
        self.lines = []
        self.event = ''
        self.site = ''
        self.date = ''
        self.round = ''
        self.white = ''
        self.black = ''
        self.eco = ''
        self.ecot = ''
        self.plycount = ''

    def add_line(self, line: str):
        """Saves lines read from each game.

        Args:
          line: A line read from the game.
        """
        self.lines.append(line)
        if line.startswith('[ECO '):
            self.eco = line
        elif line.startswith('[ECOT '):
            self.ecot = line
        elif line.startswith('[Event '):
            self.event = line
        elif line.startswith('[Site '):
            self.site = line
        elif line.startswith('[Date '):
            self.date = line
        elif line.startswith('[Round '):
            self.round = line
        elif line.startswith('[White '):
            self.white = line
        elif line.startswith('[Black '):
            self.black = line
        elif line.startswith('[PlyCount '):
            self.plycount = line


def read_games(inpgnfn: str, encoding: str='utf-8') -> List[Game]:
    """Read games from input pgn file.

    Args:
      inpgnfn: The input pgn file.
      encoding: Encoding used in reading a file. If you encounter a
        ``UnicodeDecodeError`` we can use ``ISO-8859-1`` instead of ``utf-8.``

    Returns:
      A list of Game objects.
    """
    games, current = [], None
    with open(inpgnfn, 'r', encoding=encoding) as f:
        for line in f:
            if line.startswith('[Event '):
                if current:
                    games.append(current)
                current = Game()
            current.add_line(line)
        if current:
            games.append(current)
    return games


def save_games(games: List[Game], outpgnfn: str):
    """Save the games to output file.

    Args:
      games: A list of games.
      outpgnfn: The output file.
    """
    with open(outpgnfn, 'w', encoding='utf-8') as f:
        for game in games:
            for line in game.lines:
                f.write(line)


def sort_games(inpgnfn: str, outpgnfn: str, sort_tag: str, sort_direction: str):
    """Sort and save games based on game tags.

    Args:
      inpgnfn: The input pgn file.
      sort_tag: The sorting criteria, can be event, white, black, data, etc.
      sort_direction: Direction can be hightolow or lowtohigh.
    """
    games = read_games(inpgnfn, encoding='utf-8')
    sort_value = False if sort_direction == 'lowtohigh' else True
    sort_tag = sort_tag.lower()
    if sort_tag == 'ecot':
        s = sorted(games, key=attrgetter('eco'), reverse=sort_value)
        games = sorted(s, key=attrgetter(sort_tag), reverse=sort_value)
    elif sort_tag == 'eco':
        s = sorted(games, key=attrgetter('ecot'), reverse=sort_value)
        games = sorted(s, key=attrgetter(sort_tag), reverse=sort_value)
    else:
        games = sorted(games, key=attrgetter(sort_tag), reverse=sort_value)
    save_games(games, outpgnfn)



"""Add eco, opening and variation names to the input pgn file.

The eco codes, opening and variations names are coming from
the file eco.pgn that you have to supply to pgnhelper for this to work.

eco.pgn file sources:

  * eco.pgn from `pgn-extract <https://www.cs.kent.ac.uk/people/staff/djb/pgn-extract/>`_
  * eco.pgn from `my eco repository in github <https://github.com/fsmosca/eco>`_
  * eco.pgn from `pgnhelper repository <https://github.com/fsmosca/pgnhelper>`_

Example::

  >>> import pgnhelper.eco
  >>> pgnhelper.eco.add_eco("./pgn/candidates_zurich_1953.pgn", "eco_cz.pgn", "./eco/eco.pgn")

Example output from ``eco_cz.pgn``::
  
  [Event "ct"]
  [Site "Zurich"]
  [Date "1953.??.??"]
  [Round "01"]
  [White "Szabo L"]
  [Black "Geller E"]
  [Result "0-1"]
  [ECO "A15"]
  [ECOT "E02"]
  [Opening "English"]
  [OpeningT "Catalan"]
  [Variation "Anglo-Indian"]
  [VariationT "open, 5.Qa4"]
  
  1. c4 Nf6 2. g3 e6 3. Bg2 d5 4. d4 dxc4 5. Qa4+ Nbd7 ...

Note there are ``ECOT, OpeningT, and VariationT,`` these are new tags where T refers
to ``Transposition.`` The ECO ``A15`` is the ECO based on the first 2 moves and
ECOT ``E02`` is the ECO after 12 moves.
"""


import chess.pgn


def create_eco_db(inecopgnfn: str):
    """Creats a dictionary of eco data.

    Args:
      inecopgnfn: The eco.pgn file to be converted to a dictionary.
    """
    eco_db = {}
    with open(inecopgnfn, 'r') as f:
        while True:
            game = chess.pgn.read_game(f)
            if game is None:
                break
            try:
                eco = game.headers['ECO']
            except KeyError:
                continue
            try:
                opening = game.headers['Opening']
            except KeyError:
                continue
            try:
                variation = game.headers['Variation']
            except KeyError:
                variation = None
            node = game
            node_end = node.end()
            end_board = node_end.board()
            epd = end_board.epd()
            eco_db.update({epd: {'eco': eco, 'opening': opening, 'variation': variation}})
    return eco_db


def add_eco(inpgnfn: str, outpgnfn: str, inecopgnfn: str, ply: int=4, maxply: int=24):
    """Add eco, opening and variation names to the pgn file.

    Args:
      inpgnfn: The input pgn file.
      outpgnfn: The output file.
      inecopgnfn: The eco.pgn file.
      ply: The game ply to start classifying the opening.
      maxply: The max game ply to stop classifying the opening.
    """
    eco_db = create_eco_db(inecopgnfn)
    with open(outpgnfn, 'w') as w:
        with open(inpgnfn, 'r') as f:
            while True:
                game = chess.pgn.read_game(f)
                if game is None:
                    break
                first_eco, eco_t = None, None
                first_opening, opening_t = None, None
                first_variation, variation_t = None, None
                is_first_eco = False
                first_eco_ply = 1
                for node in game.mainline():
                    board = node.board()
                    gply = board.ply()
                    epd = board.epd()
                    if gply >= ply:
                        if epd in eco_db:
                            if not is_first_eco:
                                is_first_eco = True
                                first_eco = eco_db[epd]['eco']
                                first_opening = eco_db[epd]['opening']
                                first_variation = eco_db[epd]['variation']
                                first_eco_ply = board.ply()
                            else:
                                eco_t = eco_db[epd]['eco']
                                opening_t = eco_db[epd]['opening']
                                variation_t = eco_db[epd]['variation']
                            if gply >= maxply:
                                break
                        # For first eco, if pos is not in db we will takeback 1 ply until
                        elif not is_first_eco:
                            new_board = board.copy()
                            while True:
                                new_board.pop()
                                new_epd = new_board.epd()
                                if new_epd in eco_db:
                                    is_first_eco = True
                                    first_eco = eco_db[new_epd]['eco']
                                    first_opening = eco_db[new_epd]['opening']
                                    first_variation = eco_db[new_epd]['variation']
                                    first_eco_ply = new_board.ply()
                                    break
                                if new_board.ply() <= 1:
                                    break
                                if new_board.ply() <= first_eco_ply:
                                    break
                mygame = game
                if first_eco is not None:
                    mygame.headers['ECO'] = first_eco
                    mygame.headers['Opening'] = first_opening
                    if first_variation is not None:
                        mygame.headers['Variation'] = first_variation
                if eco_t is not None:
                    mygame.headers['ECOT'] = eco_t
                    mygame.headers['OpeningT'] = opening_t
                    if variation_t is not None:
                        mygame.headers['VariationT'] = variation_t
                w.write(f'{mygame}\n\n')



from typing import Union

import pandas as pd


def expected_score(rating_a: int, rating_b: int) -> float:
    """Calculates the expected score of player_a against player_b.

    Args:
      rating_a: Rating of player_a
      rating_b: Rating of player_b
    
    Returns:
      expected score of player_a

    Example::

      >>> import pgnhelper.elo
      >>> white_elo = 2600
      >>> black_elo = 2500
      >>> score = pgnhelper.elo.expected_score(white_elo, black_elo)
      >>> score
      0.6400649998028851
    """
    rd = rating_b - rating_a
    return 1 / (1+10**(rd/400))


def add_rating_change(df: pd.DataFrame, is_rating: bool, k: int=10) -> pd.DataFrame:
    """Adds rating change columns to existing df.

    Args:
      df: A dataframe of players match records.
      is_rating: True if players have rating.
      k: The k factor.

    Returns:
      A dataframe of game records with elo rating change columns.
      [Round, White, Black, WElo, BElo, Result, Wpt, Bpt, Arm, WRChg, BRChg]
    """
    if not is_rating:
        df['WRChg'] = 0
        df['BRChg'] = 0
    else:
        df['WRChg'] = k * (df.Wpt - expected_score(df.WElo, df.BElo))
        df['BRChg'] = k * (df.Bpt - expected_score(df.BElo, df.WElo))
    return df


def get_rating(df: pd.DataFrame, p: str) -> Union[int, str]:
    """Gets the rating of player p.

    Args:
      df: A dataframe of players match records.
      p: A player name.

    Returns:
      rating of player p
    """
    dfr = df.loc[df.White == p]
    if len(dfr):
        return dfr.iloc[0]['WElo']
    dfr = df.loc[df.Black == p]
    if len(dfr):
        return dfr.iloc[0]['BElo']
    return '?'


def get_rating_change(df: pd.DataFrame, p: str, k: int=10) -> float:
    """Gets the rating change of player p.

    The given df has a rating change column for each side. This rating
    change column was calculated using a k factor with value 10. The
    k parameter if not 10 will be used to recalculate the rating change
    for the given player.

    Armageddon games are not included in the returned rating change value.

    Args:
      df: A dataframe of players match records.
      p: A player named p.
      k: The rating change k factor.

    Returns:
      rating change

    Eample 1, get the rating change of a player in the PGN file::

      >>> import pgnhelper.roundrobin
      >>> import pgnhelper.elo
      >>> df, players, is_rating = pgnhelper.roundrobin.get_pgn_data("./pgn/superbet_classic_2022_bucharest.pgn")
      >>> players
      ['Firouzja, Alireza', 'Aronian, Levon', 'So, Wesley', 'Nepomniachtchi, Ian', 'Caruana, Fabiano', 'Vachier-Lagrave, Maxime', 'Deac, Bogdan-Daniel', 'Mamedyarov, Shakhriyar', 'Dominguez Perez, Leinier', 'Rapport, Richard']
      >>> rc_levon = pgnhelper.elo.get_rating_change(df, "Aronian, Levon", k=10)
      >>> rc_levon
      9.496967974633389

    Eample 2, get the rating change of black player::

      >>> import pgnhelper.elo
      >>> white_rating = 2700
      >>> black_rating = 2600
      >>> black_point = 1
      >>> white_point = 0
      >>> expected_score = pgnhelper.elo.expected_score(black_rating, white_rating)
      >>> k = 10
      >>> rating_change = k * (black_point - expected_score)
      >>> rating_change
      6.400649998028851
    """
    dfw = df.loc[(df.White == p) & (df.Arm == 0)]
    dfb = df.loc[(df.Black == p) & (df.Arm == 0)]
    s = dfw.WRChg.sum() + dfb.BRChg.sum()
    if k != 10:
        s = s / 10
        return k * s
    return s



"""Manages command line processes.
"""


import argparse
import pgnhelper.app


def main():
    parser = argparse.ArgumentParser()

    subparser = parser.add_subparsers(dest='command')
    sort = subparser.add_parser('sort',
        help='Sort the games from the given pgn file based on the given game tags. e.g. '
        'pgnhelper sort mygames.pgn --outpgnfn out.pgn --sort-tag opening --sort-direction hightolow')
    addeco = subparser.add_parser('addeco',
        help='Add eco and ecot codes, opening and variation names to the input pgn file. '
        'The eco, opening etc. are from the given input file eco.pgn. e.g. '
        'pgnhelper addeco --inpgnfn mygames.pgn --inecopgnfn eco.pgn --outpgnfn out.pgn')
    roundrobin = subparser.add_parser('roundrobin',
        help='Generate round-robin table results from the input pgn file. '
        'The output can be html, csv and txt. e.g. '
        'pgnhelper roundrobin --inpgnfn candidates.pgn --output candidates.html')

    sort.add_argument('--inpgnfn', type=str, required=True,
        help='Write the input pgn filename, required.')
    sort.add_argument('--outpgnfn', type=str, required=True,
        help='Write the output pgn filename, required, mode=overwrite.')
    sort.add_argument('--sort-tag', type=str, required=False, default='eco',
        help='Sort the games by tag. [default=eco, value=(eco | ecot | event | date | round | white | black | site | plycount)].'
        ' e.g. --sort-tag event')
    sort.add_argument('--sort-direction', type=str, required=False, default='lowtohigh',
        help='Write the direction to sort the games. [default=lowtohigh, value=(lowtohigh | hightolow)].')
    sort.add_argument('--encoding', type=str, required=False, default='utf-8',
        help='Encoding used in reading pgn file when sorting, not required. [default=utf-8, value=(utf-8 | ISO-8859-1)]. '
        'If you encounter an error like "UnicodeDecodeError: utf-8 codec cannot decode ..." you can try, '
        '--encoding ISO-8859-1')

    addeco.add_argument('--inpgnfn', type=str, required=True,
        help='Write the input pgn filename, required.')
    addeco.add_argument('--outpgnfn', type=str, required=True,
        help='Write the output pgn filename, required, mode=overwrite.')
    addeco.add_argument('--inecopgnfn', type=str, required=True,
        help='Write the reference eco.pgn filename, required.')

    roundrobin.add_argument('--inpgnfn', type=str, required=True,
        help='Write the input pgn filename, required.')
    roundrobin.add_argument('--output', type=str, required=True,
        help='Write the output filename, required, can be .html, .csv or .txt. e.g '
        '--output tata_steel.html')
    roundrobin.add_argument('--win-point', type=float,  required=False, default=1.0,
        help='The point when the player wins, default=1.0')
    roundrobin.add_argument('--draw-point', type=float, required=False, default=0.5,
        help='The point when the player draws, default=0.5')
    roundrobin.add_argument('--armageddon-file', type=str, required=False, default=None,
        help='The armageddon pgn file, not required, default=None, if the tournament is governed by '
        'armageddon tie-break system, you need to intput the armageddon pgn file.')
    roundrobin.add_argument('--win-point-arm', type=float, required=False, default=1.0,
        help='The point when one player wins the armageddon match, not required, default=1.0')
    roundrobin.add_argument('--loss-point-arm', type=float, required=False, default=0.0,
        help='The point when one player loses the armageddon match, not required, default=0.0')
    roundrobin.add_argument('--show-max-score', action='store_true',
        help='A flag to show MaxScore column in the table, can be useful when scoring is not standard.')
    roundrobin.add_argument('--table-color', type=str, required=False, default='blue_light',
        help='Write table color not required. [default="blue_light" value=("yellow_light", '
        '"grey_light", "orange_light", "green_light", "red_light", "yellow_dark", '
        '"grey_dark", "blue_dark", "orange_dark", "green_dark", "red_dark")]')

    parser.add_argument('-v', '--version', action='version', version=f'{pgnhelper.__version__}')

    args = parser.parse_args()

    g = None
    if args.command == 'sort':
        g = pgnhelper.app.PgnHelper(
            args.command,
            inpgnfn=args.inpgnfn,
            outpgnfn=args.outpgnfn,
            sort_tag=args.sort_tag,
            sort_direction=args.sort_direction,
            encoding=args.encoding
        )
    elif args.command == 'addeco':
        g = pgnhelper.app.PgnHelper(
            args.command,
            inpgnfn=args.inpgnfn,
            outpgnfn=args.outpgnfn,
            inecopgnfn=args.inecopgnfn            
        )
    elif args.command == 'roundrobin':
        g = pgnhelper.app.PgnHelper(
            args.command,
            inpgnfn=args.inpgnfn,
            output=args.output,
            winpoint = args.win_point,
            drawpoint = args.draw_point,
            tablecolor = args.table_color,
            armageddonfile = args.armageddon_file,
            winpointarm = args.win_point_arm,
            losspointarm = args.loss_point_arm,
            showmaxscore=args.show_max_score
        )
    
    if g is not None:
        g.start()


if __name__ == '__main__':
    main()



"""Helps build the round-robin result table.
"""


from typing import List
import pandas as pd


def get_encounter_score(df: pd.DataFrame, p: str, op: str, winpoint: float=1.0,
        drawpoint: float=0.5, winpointarm: float=1.0,
        losspointarm: float=0.0) -> List[float]:
    """Calculates the scores between the player p and op.

    Args:
      df: A pandas dataframe containing players match results.
      p: A player name.
      op: Opponent of player p.
      winpoint: The point when player wins.
      drawpoint: The point when player draws.
      winpointarm: The point when player wins in armageddon game.
      losspointarm: The point when player loses in armageddon game.

    Returns:
      A list of score for p and op, score[<p score>, <op score>].
    """
    is_arm = True if 1 in df.Arm.unique() else False
    score = [0, 0]
    dfw = df[(df.White == p) & (df.Black == op)]
    dfb = df[(df.Black == p) & (df.White == op)]

    # If a game is normal or no tie-break points involved.
    if not is_arm:
        # Update the score when p is white.    
        for i in range(len(dfw)):
            v = dfw.iloc[i]['Result']
            if v == '1-0':
                score[0] += 1.0 * winpoint
                score[1] += 0.0
            elif v == '0-1':
                score[0] += 0.0
                score[1] += 1.0 * winpoint
            elif v == '1/2-1/2':
                score[0] += drawpoint
                score[1] += drawpoint

        # Update the score when p is black.
        for i in range(len(dfb)):
            v = dfb.iloc[i]['Result']
            if v == '1-0':
                score[0] += 0.0
                score[1] += 1.0 * winpoint
            elif v == '0-1':
                score[0] += 1.0 * winpoint
                score[1] += 0.0
            elif v == '1/2-1/2':
                score[0] += drawpoint
                score[1] += drawpoint

    # Else if a game has armageddon tie-break.
    else:
        # A. In norway chess, in normal game, the winner will get 3.0 points, the loser will get 0.0 point.

        # 1. Get the score from normal game with 1-0 and 0-1 result where the first player is white.
        # We don't include the draw results because we will use the armageddon tie-break system.
        dfw_normal = df[(df.White == p) & (df.Black == op) & (df.Arm == 0) & ((df.Result == '1-0') | (df.Result == '0-1'))]
        for i in range(len(dfw_normal)):
            r = dfw_normal.iloc[i]['Result']
            if r == '1-0':
                score[0] += 1.0 * winpoint
                score[1] += 0.0
            elif r == '0-1':
                score[0] += 0.0
                score[1] += 1.0 * winpoint

        # 2. Get the score from normal game with 1-0 and 0-1 result where the first player is black.
        # We don't include the draw results because we will use the armageddon tie-break system.
        dfb_normal = df[(df.Black == p) & (df.White == op) & (df.Arm == 0) & ((df.Result == '1-0') | (df.Result == '0-1'))]
        for i in range(len(dfb_normal)):
            r = dfb_normal.iloc[i]['Result']
            if r == '0-1':
                score[0] += 1.0 * winpoint
                score[1] += 0.0
            elif r == '1-0':
                score[0] += 0.0
                score[1] += 1.0 * winpoint

        # B. In norway chess, in armageddon game, the winner will get 1.5 points, the loser will get 1.0 point.
        # The armageddon game is done if the normal game resulted in a draw. The point in a normal game that
        # resulted in a draw is discarded.

        # 1. Get the score from armageddon game where the first player is white.
        dfw_arm = df[(df.White == p) & (df.Black == op) & (df.Arm == 1)]
        for i in range(len(dfw_arm)):
            r = dfw_arm.iloc[i]['Result']
            if r == '1-0':
                score[0] += 1.0 * winpointarm
                score[1] += 1.0 * losspointarm
            elif r == '0-1':
                score[0] += 1.0 * losspointarm
                score[1] += 1.0 * winpointarm
            elif r == '1/2-1/2':
                score[0] += 1.0 * losspointarm
                score[1] += 1.0 * winpointarm  # black wins
        # 2. Get the score from armageddon game where the first player is black.
        dfb_arm = df[(df.Black == p) & (df.White == op) & (df.Arm == 1)]
        for i in range(len(dfb_arm)):
            r = dfb_arm.iloc[i]['Result']
            if r == '0-1':
                score[0] += 1.0 * winpointarm
                score[1] += 1.0 * losspointarm
            elif r == '1-0':
                score[0] += 1.0 * losspointarm
                score[1] += 1.0 * winpointarm
            elif r == '1/2-1/2':
                score[0] += 1.0 * winpointarm  # black wins
                score[1] += 1.0 * losspointarm
    return score



"""Generates a round-robin result table.

It reads the input pgn file and generates a dataframe of round-robin table.
It also add columns for tie-break scores for tied players.

Typical tie-break system that can be applied to a round-robin tournament
according to FIDE.

13.16.2. Individual Round-Robin Tournaments:
    * Direct encounter
    * The greater number of wins, including forfeits
    * Sonneborn-Berger
    * Koya System (to be implemented in pgnhelper)

https://handbook.fide.com/files/handbook/C02Standards.pdf
"""


from pathlib import Path
from typing import List, Tuple, Optional

import chess.pgn
import pandas as pd
from pretty_html_table import build_table

import pgnhelper.tiebreak
import pgnhelper.utility
import pgnhelper.elo


def get_pgn_data(fn: str, is_arm: bool=False, k: int=10) -> Tuple[pd.DataFrame, List, bool]:
    """Converts games to dataframe.

    Args:
      fn: A pgn file.
      is_arm: If pgn file has armageddon games.
      k: The rating change k factor.

    Returns:
      [records, players, is_rating]
    """
    data = []
    players = []
    rating_cnt = 0
    with open(fn, 'r') as f:
        while True:
            game = chess.pgn.read_game(f)
            if game is None:
                break
            round = game.headers['Round']
            white = game.headers['White']
            black = game.headers['Black']
            result = game.headers['Result']
            players.append(white)
            players.append(black)
            welo = game.headers.get('WhiteElo', '?')
            belo = game.headers.get('BlackElo', '?')
            welo = '?' if welo == '' else welo
            belo = '?' if belo == '' else belo
            if welo != '?':
                rating_cnt += 1
                welo = int(welo)
            if belo != '?':
                rating_cnt += 1
                belo = int(belo)
            if result == '1-0':
                wpt = 1.0
                bpt = 0.0
            elif result == '0-1':
                wpt = 0.0
                bpt = 1.0
            elif result == '1/2-1/2':
                wpt = 0.5
                bpt = 0.5            
            data.append([round, white, black, welo, belo, result, wpt, bpt, 1 if is_arm else 0])
    df = pd.DataFrame(data,
            columns=['Round', 'White', 'Black', 'WElo', 'BElo', 'Result', 'Wpt', 'Bpt', 'Arm'])
    df = pgnhelper.elo.add_rating_change(df, rating_cnt > 1, k)
    return df, list(set(players)), rating_cnt > 0


def games_per_encounter(result_df: pd.DataFrame, ranking_df: pd.DataFrame) -> int:
    """Count the number of games per encounter excluding armageddon.

    Args:
      result_df: A dataframe of game result records.
      ranking_df: A dataframe of player rankings.

    Returns:
      The number of games per encounter.
    """
    players = list(ranking_df.Name)
    for p in players:
        for m in players:
            if p == m:
                continue
            dfw = result_df.loc[(result_df.White == p) & (result_df.Black == m)
                    & (result_df.Arm == 0)]
            dfb = result_df.loc[(result_df.Black == p) & (result_df.White == m)
                    & (result_df.Arm == 0)]
            return len(dfw) + len(dfb)
    return 0


def player_ranking(df: pd.DataFrame, players: List, is_rating: bool,
        winpoint: float, drawpoint: float, winpointarm:float=1.5,
        losspointarm: float=1.0) -> pd.DataFrame:
    """Generates a dataframe of player ranking.

    Args:
      df: A dataframe of player match records.
      players: A list of unique players names.
      is_rating: If players have rating.
      winpoint: The point when players wins.
      drawpoint: The point when player draws.

    Returns:
      A pandas dataframe of players ranking.
    """
    is_arm = True if 1 in df['Arm'].unique() else False
    data_p = []
    for p in players:       
        if not is_arm:
            df_w = df[df.White == p]
            df_b = df[df.Black == p]
            score_w = len(df_w[df_w.Result == '1-0']) * winpoint
            score_w += len(df_w[df_w.Result == '1/2-1/2']) * drawpoint
            score_b = len(df_b[df_b.Result == '0-1']) * winpoint
            score_b += len(df_b[df_b.Result == '1/2-1/2']) * drawpoint
            final_score = score_w + score_b
        else:
            # Used to get the number of games for normal games.
            df_w = df[(df.White == p) & (df.Arm == 0)]
            df_b = df[(df.Black == p) & (df.Arm == 0)]

            # Get the score for normal games without draws.
            df_w_normal = df.loc[(df.White == p) & (df.Arm == 0) & (df.Result != '1/2-1/2')]
            df_b_normal = df.loc[(df.Black == p) & (df.Arm == 0) & (df.Result != '1/2-1/2')]
            score_w_n = len(df_w_normal.loc[df_w_normal.Result == '1-0']) * winpoint
            score_b_n = len(df_b_normal.loc[df_b_normal.Result == '0-1']) * winpoint

            # Get score for arm games.
            df_w_arm = df.loc[(df.White == p) & (df.Arm == 1)]
            df_b_arm = df.loc[(df.Black == p) & (df.Arm == 1)]

            # wins
            score_w_aw = len(df_w_arm.loc[df_w_arm.Result == '1-0']) * winpointarm
            score_b_aw = len(df_b_arm.loc[df_b_arm.Result == '0-1']) * winpointarm
            score_b_aw += len(df_b_arm.loc[df_b_arm.Result == '1/2-1/2']) * winpointarm

            # loses
            score_w_al = len(df_w_arm.loc[df_w_arm.Result == '0-1']) * losspointarm
            score_w_al += len(df_w_arm.loc[df_w_arm.Result == '1/2-1/2']) * losspointarm
            score_b_al = len(df_b_arm.loc[(df_b_arm.Result == '1-0')]) * losspointarm

            final_score = (score_w_n + score_b_n + score_w_aw + score_b_aw
                    + score_w_al + score_b_al)

        if is_rating:
            rating = df_w.WElo.iloc[0]  # all players have at least played a game with white
            data_p.append([p, rating, len(df_w) + len(df_b), final_score])
        else:
            data_p.append([p, len(df_w) + len(df_b), final_score])
    if is_rating:
        df_score = pd.DataFrame(data_p,
                columns=['Name', 'Rating', 'Games', 'Score'])
    else:
        df_score = pd.DataFrame(data_p, columns=['Name', 'Games', 'Score'])
    df_score = df_score.sort_values(by=['Score', 'Name'],
            ascending=[False, True])
    df_score = df_score.reset_index(drop=True)
    return df_score


def save_roundrobin_table(df: pd.DataFrame,
        outputfn: str, tablecolor: str='blue_light') -> None:
    """Save the round-robin result table.

    The output can be a csv, txt and html.

    Args:
      df: A dataframe of players match records.
      outputfn: The output file.
      tablecolor: The round-robin table color.

    Example, save the round-robin table in html file::

      import pgnhelper.roundrobin

      df = pgnhelper.roundrobin.round_robin('airthings.pgn')
      pgnhelper.roundrobin.save_roundrobin_table(df, 'airthings.html')
    """
    ext = Path(outputfn).suffix
    if ext == '.html':
        html_table = build_table(df, tablecolor,
            font_size='medium',
            text_align='center',
            font_family='Calibri, Verdana, Tahoma, Georgia, serif, arial')
        with open(outputfn, 'w') as f:
            f.write(html_table)
    elif ext == '.csv':
        df.to_csv(outputfn, index=False)
    else:
        df.to_string(outputfn, index=False)


def round_robin(fn: str, winpoint: float=1.0, drawpoint: float=0.5,
        armageddonfile: Optional[str]=None, winpointarm: float=1.0,
        losspointarm: float=0.0, showmaxscore: bool=False) -> pd.DataFrame:
    """Generates a round-robin result table.

    Args:
      fn: The input pgn file.
      winpoint: The point when player wins.
      drawpoint: The point when player draws.
      armageddonfile: The pgn file from armageddon games.
      winpointarm: The point when player wins in an armageddon game.
      losspointarm: The point when player loses in an armageddon game.
      showmaxscore: Show the column maxscore in the generated table.

    Returns:
      A pandas dataframe of round-robin result table.
    """
    dfall = []
    dfn, players, is_rating = get_pgn_data(fn, is_arm=False)
    dfall.append(dfn)
    if armageddonfile is not None:
        dfa, _, _ = get_pgn_data(armageddonfile, is_arm=True)
        dfall.append(dfa)
    df = pd.concat(dfall, ignore_index=False)
    is_arm = True if 1 in df.Arm.unique() else False

    # 1. Create a dataframe of player ranking.
    df_score = player_ranking(df, players, is_rating, winpoint,
            drawpoint, winpointarm, losspointarm)
    gpe = games_per_encounter(df, df_score)

    # 1.1 Apply Direct Encounter tie-break
    df_de = pgnhelper.tiebreak.direct_encounter(df, df_score, winpoint,
            drawpoint, winpointarm, losspointarm)
    df_de = df_de.sort_values(by=['Score', 'DE', 'Name'],
            ascending=[False, False, True])
    df_de = df_de.reset_index(drop=True)

    # 1.2 Apply Number of Wins tie-break
    df_wins = pgnhelper.tiebreak.num_wins(df, df_de)
    df_wins = df_wins.sort_values(by=['Score', 'DE', 'Wins', 'Name'],
            ascending=[False, False, False, True])
    df_wins = df_wins.reset_index(drop=True)  

    # 1.3 Apply Sonneborn-Berger tie-break
    if not is_arm:
        df_sb = pgnhelper.tiebreak.sonneborn_berger(df, df_wins, gpe=gpe,
                winpoint=1.0, drawpoint=0.5)
        df_sb = df_sb.sort_values(by=['Score', 'DE', 'Wins', 'SB', 'Name'],
                ascending=[False, False, False, False, True])
        df_sb = df_sb.reset_index(drop=True)
    else:
        df_sb = df_wins.copy()

    # 2. Build a round-robin dataframe.
    if is_rating:
        # Add rating change.
        rc = []
        for p in list(df_sb.Name):
            r = pgnhelper.elo.get_rating_change(df, p, k=10)
            rc.append(round(r, 2))
        data_rr = {'Name': df_sb.Name, 'Rating': df_sb.Rating, 'RChg': rc}
    else:
        data_rr = {'Name': df_sb.Name.unique()}
    cnt = 1
    for p in df_sb.Name.unique():
        data_v = []
        for op in df_sb.Name.unique():
            if p == op:
                v = 'x'
            else:
                score = pgnhelper.utility.get_encounter_score(df, p, op,
                        winpoint, drawpoint, winpointarm, losspointarm)
                v = score[1]  # use the score of op only
            data_v.append(v)
        data_rr.update({cnt: data_v})
        cnt += 1
    df_rr = pd.DataFrame(data_rr)

    # 3. Add other columns at the end.
    df_rr['Games'] = df_sb['Games']
    df_rr['Score'] = df_sb['Score']
    max_score = df_sb['Games'] * winpoint
    if showmaxscore:
        df_rr['MaxScore'] = max_score
    df_rr['Score%'] = 100 * df_sb['Score'] / max_score
    df_rr['Score%'] = df_rr['Score%'].round(2)
    df_rr['DE'] = df_sb['DE'].round(2)
    df_rr['Wins'] = df_sb['Wins'].round(0)
    if not is_arm:
        df_rr['SB'] = df_sb['SB'].round(2)

    # 4. Insert rank column at first column.
    df_rr.insert(loc=0, column='Rank', value=range(1, len(df_rr) + 1))
    return df_rr



"""Manages job requests from user through the command line.
"""


from typing import Optional

import pgnhelper.eco
import pgnhelper.sort
import pgnhelper.roundrobin


class PgnHelper:
    """Manages user options to execute the job.

    Attributes:
      job: The kind of job to be done, sort, addeco and roundrobin.
      inpgnfn: The input pgn file or path and filename.
      outpgnfn: The output pgn file or path and filename.
      inecopgnfn: The eco.pgn that will be used in addeco job.
      sort_tag: Used in sorting games. 
      sort_direction: The output sorting ordering, lowtohigh or hightolow.       
    """

    def __init__(self, job: str, inpgnfn: Optional[str]=None,
            outpgnfn: Optional[str]=None, inecopgnfn: Optional[str]=None,
            sort_tag: str='eco', sort_direction: str='lowtohigh',
            output: Optional[str]=None, winpoint: float=1.0,
            drawpoint: float=0.5, tablecolor: str='blue_light',
            encoding: str='utf-8', armageddonfile: Optional[str]=None,
            winpointarm: float=1.0, losspointarm: float=0.0,
            showmaxscore: bool=False):
        self.job = job
        self.inpgnfn = inpgnfn
        self.inecopgnfn = inecopgnfn
        self.outpgnfn = outpgnfn
        self.sort_tag = sort_tag
        self.sort_direction = sort_direction
        self.output = output
        self.winpoint = winpoint
        self.drawpoint = drawpoint
        self.tablecolor = tablecolor
        self.encoding = encoding
        self.armageddonfile = armageddonfile
        self.winpointarm = winpointarm
        self.losspointarm = losspointarm
        self.showmaxscore = showmaxscore

    def start(self):
        """Run the type of job to be done.

        It will sort the games, add eco, opening and variation names to
        the games or generate a round-robin result table.
        """
        if self.job == 'sort':
            pgnhelper.sort.sort_games(self.inpgnfn, self.outpgnfn,
                    self.sort_tag, self.sort_direction)
        elif self.job == 'addeco':
            pgnhelper.eco.add_eco(self.inpgnfn, self.outpgnfn,
                    self.inecopgnfn, ply=4, maxply=24)
        elif self.job == 'roundrobin':
            df = pgnhelper.roundrobin.round_robin(
                self.inpgnfn,
                winpoint=self.winpoint,
                drawpoint=self.drawpoint,
                armageddonfile=self.armageddonfile,
                winpointarm=self.winpointarm,
                losspointarm=self.losspointarm,
                showmaxscore=self.showmaxscore)
            pgnhelper.roundrobin.save_roundrobin_table(df,
                    self.output, self.tablecolor)
